/*======================================================================

  $Id: mtd.c,v 1.8 1999/10/25 14:55:46 dwmw2 Exp $

    A general driver for accessing Memory Technology Devices

    This driver provides the equivalent of /dev/mem for a MTD
    device. It includes a character device - a block device is
    implemented by mtdblock.c
    
======================================================================*/

#include <linux/module.h>
#include <asm/uaccess.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/major.h>
#include <linux/fs.h>
#include <linux/ioctl.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/segment.h>
#include <stdarg.h>
#include <linux/init.h>
#ifdef CONFIG_PROC_FS
#include <linux/proc_fs.h>
#endif

#include <linux/mtd/mtd.h>

/* Major device #'s for memory device */

static struct mtd_info *mtd_table[MAX_MTD_DEVICES];



static loff_t mtd_lseek (struct file *file, loff_t offset, int orig);
static int mtd_ioctl(struct inode *inode, struct file *file,
		     u_int cmd, u_long arg);
static ssize_t mtd_read(struct file *file, char *buf, size_t count,
		     loff_t *ppos);
static ssize_t mtd_write(struct file *file, const char *buf, size_t count,
			 loff_t *ppos);
static int mtd_open(struct inode *inode, struct file *file);
static int mtd_close(struct inode *inode,
		     struct file *file);

static struct file_operations mtd_fops = {
	mtd_lseek,     	/* lseek */
	mtd_read,	/* read */
	mtd_write, 	/* write */
	NULL,		/* readdir */
	NULL,		/* poll */
	mtd_ioctl,	/* ioctl */
	NULL,		/* mmap */
	mtd_open,	/* open */
	NULL,		/* flush */
	mtd_close,	/* release */
	NULL		/* fsync */
};

static struct mtd_notifier *mtd_notifiers = NULL;

static loff_t mtd_lseek (struct file *file, loff_t offset, int orig)
{
	int devnum = MINOR(file->f_dentry->d_inode->i_rdev) >> 1;
	struct mtd_info *mtd=mtd_table[devnum];

	if (devnum >=MAX_MTD_DEVICES || !mtd) 
		return -ENODEV;

	switch (orig) {
	case 0: 
		/* SEEK_SET */
		file->f_pos = offset;
		break;
	case 1: 
		/* SEEK_CUR */
		file->f_pos += offset;
		break;
	case 2:
		/* SEEK_END */
		file->f_pos =mtd->size + offset;
		break;
	default: 
		return -EINVAL;
	}

	if (file->f_pos < 0) 
		file->f_pos = 0;
	else if (file->f_pos >= mtd->size)
		file->f_pos = mtd->size - 1;

	return file->f_pos;
}

static int mtd_open(struct inode *inode, struct file *file)
{
	int minor = MINOR(inode->i_rdev);
	int devnum = minor >> 1;
	struct mtd_info *mtd=mtd_table[devnum];	
	
	if (devnum >= MAX_MTD_DEVICES || !mtd)
		return -ENODEV;
	
	if (file->f_mode & 2)
	{
		/* We're trying to open it with write access. */
		/* Fail if it's read-only, or if we're using the ro device */
		if ((minor & 1) || !(mtd->flags & MTD_WRITEABLE))
			return -EACCES;
	}

	/* OK. We're going to open it now. */

	if (mtd->module)
	  __MOD_INC_USE_COUNT(mtd->module);

	return 0;
} /* mtd_open */

/*====================================================================*/

static int mtd_close(struct inode *inode,
				 struct file *file)
{
	int minor = MINOR(inode->i_rdev);
	struct mtd_info *mtd;
	int devnum = minor >> 1;

	if (devnum >= MAX_MTD_DEVICES)
		return -ENODEV;

	mtd = mtd_table[devnum];
	
	if (!mtd) return -ENODEV;
	MTD_SYNC(mtd);

	if (mtd->module)
		__MOD_DEC_USE_COUNT(mtd->module);

	return 0;
} /* mtd_close */


static ssize_t mtd_read(struct file *file, char *buf, size_t count,
			loff_t *ppos)
{
	int devnum = MINOR(file->f_dentry->d_inode->i_rdev) >> 1;
	struct mtd_info *mtd = mtd_table[devnum];
	u_char *mtdbuf;
	size_t retlen=0;
	u_long count1;
	int ret=0;

	if (!mtd) return -ENODEV;

	if (*ppos + count > mtd->size)
		count = mtd->size - *ppos;

	if (!count)
		printk("Count is zero.\n");

	count1 = count;

	while (count && !ret)
	{
		if (!mtd->point || (ret = MTD_POINT(mtd, *ppos, count, &retlen, &mtdbuf)) != 0)
		{
			/* mtd->point() failed; use mtd->read instead */
		
			ret = MTD_READ(mtd, *ppos, count, &retlen, buf);
			if (retlen)
			{
				*ppos += retlen;
				return retlen;
			}
			else 
				return ret;
		}

		if (!retlen)
		{
			printk(KERN_NOTICE "mtd_read: device returned no error, but no data either!\n");
			ret = -EIO;
			break;
		}
		
		if (copy_to_user(buf,mtdbuf,retlen) < 0) 
			ret = -EFAULT;
		else
		{
			count -= retlen;
			*ppos += retlen;
		}
		MTD_UNPOINT(mtd, mtdbuf);
	}
	
	if (count < count1)
		return count1-count;
	else 
		return ret;
} /* mtd_read */


static ssize_t mtd_write(struct file *file, const char *buf, size_t count,
			 loff_t *ppos)
{
	int devnum = MINOR(file->f_dentry->d_inode->i_rdev) >> 1;
	struct mtd_info *mtd = mtd_table[devnum];
	u_long count1;
	size_t retlen;
	int ret=0;
	
	if (!mtd) return -ENODEV;

	if (*ppos == mtd->size)
		return -ENOSPC;
	
	if (*ppos + count > mtd->size)
		count = mtd->size - *ppos;
	
	count1 = count;
	
	while (count)
	{
		ret = (*(mtd->write))(mtd, *ppos, count, &retlen, buf);

		if (ret)
			break;

		if (!retlen)
		{
			printk(KERN_NOTICE "mtd_read: device returned no error, but no data either!\n");
			ret = -EIO;
			break;
		}

		count -= retlen;
	}

	if (count < count1)
	{
		*ppos += count1-count;
		return count1-count;
	}
	else 
		return ret;
} /* mtd_write */

/*======================================================================

    IOCTL calls for getting device parameters.

======================================================================*/

static int mtd_ioctl(struct inode *inode, struct file *file,
		       u_int cmd, u_long arg)
{
	int devnum = MINOR(inode->i_rdev) >> 1;
	struct mtd_info *mtd=mtd_table[devnum];
	int ret = 0;
	u_long size;
	

	if (devnum >= MAX_MTD_DEVICES || !mtd)
		return -ENODEV;
	
	size = (cmd & IOCSIZE_MASK) >> IOCSIZE_SHIFT;
	if (cmd & IOC_IN) {
		ret = verify_area(VERIFY_READ, (char *)arg, size);
		if (ret) return ret;
	}
	if (cmd & IOC_OUT) {
		ret = verify_area(VERIFY_WRITE, (char *)arg, size);
		if (ret) return ret;
	}
	
	switch (cmd) {
	case MEMGETINFO:
		copy_to_user((struct mtd_info *)arg, mtd,
			     sizeof(struct mtd_info_user));
		break;
	case MEMERASE:
	{
		struct erase_info *erase=kmalloc(sizeof(struct erase_info),GFP_KERNEL);
		if (!erase)
			ret = -ENOMEM;
		else
		{
			memset (erase,0,sizeof(struct erase_info));
			copy_from_user(&erase->addr, (u_long *)arg,
				       2 * sizeof(u_long));
			erase->mtd = mtd;
			ret = MTD_ERASE(mtd, erase);
		}
		break;
	}

	case MEMWRITEOOB:
	{
		struct mtd_oob_buf buf;
		void *databuf;
		ssize_t retlen;
		
		copy_from_user(&buf, (struct mtd_oob_buf *)arg, sizeof(struct mtd_oob_buf));
		
		if (buf.length > 0x4096)
			return -EINVAL;

		if (!mtd->write_oob)
			ret = -EOPNOTSUPP;
		else
			ret = verify_area(VERIFY_READ, (char *)buf.ptr, buf.length);

		if (ret)
			return ret;

		databuf = kmalloc(buf.length, GFP_KERNEL);
		if (!databuf)
			return -ENOMEM;
		
		copy_from_user(databuf, buf.ptr, buf.length);

		ret = (mtd->write_oob)(mtd, buf.start, buf.length, &retlen, databuf);

		copy_to_user((void *)arg + sizeof(loff_t), &retlen, sizeof(ssize_t));

		kfree(databuf);
		break;

	}

	case MEMREADOOB:
	{
		struct mtd_oob_buf buf;
		void *databuf;
		ssize_t retlen;

		copy_from_user(&buf, (struct mtd_oob_buf *)arg, sizeof(struct mtd_oob_buf));
		
		if (buf.length > 0x4096)
			return -EINVAL;

		if (!mtd->read_oob)
			ret = -EOPNOTSUPP;
		else
			ret = verify_area(VERIFY_WRITE, (char *)buf.ptr, buf.length);

		if (ret)
			return ret;

		databuf = kmalloc(buf.length, GFP_KERNEL);
		if (!databuf)
			return -ENOMEM;
		
		ret = (mtd->read_oob)(mtd, buf.start, buf.length, &retlen, databuf);

		copy_to_user((void *)arg + sizeof(loff_t), &retlen, sizeof(ssize_t));

		if (retlen)
			copy_to_user(buf.ptr, databuf, retlen);

		kfree(databuf);
		break;
	}
			     
			     
		
		

	default:
	  printk("Invalid ioctl %x (MEMGETINFO = %x)\n",cmd, MEMGETINFO);
		ret = -EINVAL;
	}
	
	return ret;
} /* memory_ioctl */

/*====================================================================*/

int add_mtd_device(struct mtd_info *mtd)
{
	int i;

	for (i=0; i< MAX_MTD_DEVICES; i++)
		if (!mtd_table[i])
		{
			struct mtd_notifier *not=mtd_notifiers;

			mtd_table[i]=mtd;
			
#ifdef DEBUG
			printk(KERN_NOTICE "mtd: Giving out device %d to %s\n",i, mtd->name);
#endif
			while (not)
			{
				(*(not->func))(mtd,i);
				not = not->next;
			}
			MOD_INC_USE_COUNT;
			return 0;
		}
	
	return 1;
}

struct mtd_info *get_mtd_device(u_char num)
{
	return mtd_table[num];
}

int del_mtd_device (struct mtd_info *mtd)
{
	struct mtd_notifier *not=mtd_notifiers;
	int i;
	
	for (i=0; i < MAX_MTD_DEVICES; i++)
	{
		if (mtd_table[i] == mtd)
		{
			while (not)
			{
				(*(not->func))(NULL,i);
				not=not->next;
			}
			mtd_table[i] = NULL;
			MOD_DEC_USE_COUNT;

			return 0;
		}
	}

	return 1;
}



void register_mtd_notifier (struct mtd_notifier *new)
{
	new->next = mtd_notifiers;
	mtd_notifiers = new;
	MOD_INC_USE_COUNT;
}



int unregister_mtd_notifier (struct mtd_notifier *old)
{
	struct mtd_notifier **prev = &mtd_notifiers;
	struct mtd_notifier *cur;

	while ((cur = *prev))
	{
		if (cur == old)
		{
			*prev = cur->next;
			MOD_DEC_USE_COUNT;
			return 0;
		}
		prev = &cur->next;
	}
	return 1;
}
	

/*====================================================================*/
/* /proc/mtd support */

#ifdef CONFIG_PROC_FS

struct proc_dir_entry *proc_mtd;

static inline int mtd_proc_info (char *buf, int i)
{
	struct mtd_info *this = get_mtd_device(i);

	if (!this) 
		return 0;

	return sprintf(buf, "mtd%d: %8.8lx \"%s\"\n", i, this->size, 
		       this->name);
}

static int mtd_read_proc ( char *page, char **start, off_t off, 
		       int count, int *eof, void *data)
{
	int len = 0, l, i;
        off_t   begin = 0;
      
        for (i=0; i< MAX_MTD_DEVICES; i++) {

                l = mtd_proc_info(page + len, i);
                len += l;
                if (len+begin > off+count)
                        goto done;
                if (len+begin < off) {
                        begin += len;
                        len = 0;
                }
        }

        *eof = 1;
done:
        if (off >= len+begin)
                return 0;
        *start = page + (begin-off);
        return ((count < begin+len-off) ? count : begin+len-off);
}
#endif

/*====================================================================*/

#if LINUX_VERSION_CODE < 0x20300
#ifdef MODULE
#define init_mtd init_module
#define cleanup_mtd cleanup_module
#endif
#define __exit
#endif


static int __init init_mtd(void)
{
	int i;
	
	for (i=0; i<MAX_MTD_DEVICES; i++)
		mtd_table[i]=NULL;

	if (register_chrdev(MTD_CHAR_MAJOR,"mtd",&mtd_fops))
	{
		printk(KERN_NOTICE "Can't allocate major number %d for Memory Technology Devices.\n",
		       MTD_CHAR_MAJOR);
		return EAGAIN;
	}
#ifdef CONFIG_PROC_FS
	if ((proc_mtd = create_proc_entry( "mtd", 0, 0 )))
	  proc_mtd->read_proc = mtd_read_proc;
#endif
	return 0;
}

static void __exit cleanup_mtd(void)
{
	unregister_chrdev(MTD_CHAR_MAJOR, "mtd");
#ifdef CONFIG_PROC_FS
	if (proc_mtd)
	  remove_proc_entry( "mtd", 0);
#endif
}




#if LINUX_VERSION_CODE < 0x20300
#ifndef MODULE
__initfunc(init_mtd);
#endif
#else
module_init(init_mtd);
module_exit(cleanup_mtd);
#endif
